*suppress log once program has run error free;
options nonotes nosource nosource2 errors=0;

*keep a record of start time to calculate runtime;
%let startime = %sysfunc(time(),time8.);
**********************************************************;
*Filename:    A GA for Data Reduction.sas                 ;
*Author:      Lisa Henley                                 ;
*Description: Program builds a genetic algorithm which is ;
*             used to reduce a large dataset to a more    ;
*             manageable level.  The 'best' variables are ;
*             selected using the strength of  their       ;
*			  association with a dependent variable.      ;
**********************************************************;
*set up a libname to point to data to be analysed;
libname dset1 "" ;

*set control parameters these are run multiple times to ensure stability;
*and can be adjusted as per user-needs;
%let initpop = 200;
%let reps = 100; *half of initpop as binary tournament is selection method;
%let topcount = 80 ; *amount selected for crossover;
%let elite = 2 ; *number retained from one generation to another without crossover;  
%let crossreps = 39 ; *(topcount - (elite)) / 2 (tournaments);
%let maxit = 500; *maximum number of iterations;
%let stable = 20;  *number of 'same' iterations before solution judged stable;
%let selection = forward; *type of regression - forward, backward, stepwise;
%let sle = .15; *significance level for variable entry;
%let sls = .20; *significance level for variable to stay;

*for each different dataset;
%let data = rawdata; *the name of the dataset to be analysed;
%let outvars = 6; *number of variables in dataset to be analysed;

*get local copy of dataset;
data mydata;
	set dset1.&data;
run;

*create the initial population with randomly generated alleles;
data initpop (drop = i temp);
	array allele {&outvars } 3 ;
	do until (_n_ = %eval(&initpop + 1)) ;
		do i = 1 to &outvars ;
			temp=uniform(1);
			if temp >=0.5 then allele[i] = 1;
			else allele[i]=0;
		end;
		*make a chromosome of the alleles;
		chromosome = catt(of allele1-allele&outvars );
		output;
		_n_+1;
	end;
run;

*Want to select x% for crossover;
*sample without replacement and do binary tournament selection for;
*whole population;

*build dataset of samples.  The sample size is two because the ;
*selection method is binary tournament, so each replicate has ;
*two members;
proc surveyselect data=initpop method = SRS rep = &reps 
                         sampsize = 2 seed = 12345 out = testfit noprint;
  id _all_;
run;

*get variable names from dataset being analysed;
proc contents data=mydata
	out=var_names(keep=name) noprint;
run;

*tidy up - only keep the independent variable names;
data var_names;
	set var_names;
	*code here will depend on user data;
	*only variables to be included in analysis should;
	*remain here;
run;

*create a variable in sample dataset which records which partner in ;
*tournament each model is i.e. each sample (tournament) has two ;
*competitors (models) and they are numbered 1 or 2;
data testfit;
	set testfit;
	by replicate;
	if first.replicate then model=1;
	else model=2;
run;

*keep the workspace tidy for successive runs;
*before starting the loop check and delete elite dataset if necessary;
*also delete next gen dataset.  The elite dataset is the option to ;
*keep the top n performers without crossover ;

%macro cleanup(dname);
%if %sysfunc(exist(&dname)) %then %do;
    proc datasets library=work;
		delete &dname;
	quit;
	run;
%end;
%mend;

%cleanup(elite);
%cleanup(nextgen);
%cleanup(oldies);


*transpose sample for merge;
%macro fitness(count);

*delete tournament dataset before you start if it exists;
%if %sysfunc(exist(tournament)) %then %do;
	proc datasets library=work;
		delete tournament;
		quit;
	run;
%end;

*for each of the tournaments;
%do i = 1 %to &count;

	*transpose the sample for ease of working;
	proc transpose data=testfit out=testfitturn;
		where replicate =  &i;
	run;

	/*only keep variable names where the allele (col1 or col2) is flagged 1
	store these kept names in variables 'mod1' and 'mod2'.  If an allele has
	value 1 then that variable will be in the model, if it is 0 then that
	variable won't be in the model*/
	data modelnames;
		merge testfitturn (where = (_name_ ne 'Replicate' 
			and _name_ ne 'model' and _name_ ne 'chromosome')) var_names;
		if col1 = 1 then mod1 = name;
		if col2 = 1 then mod2 = name;
	run;

	*compress these two variables into a string and store in a macro;
	*variable for use in next step;
	proc sql noprint;
		select compress(mod1) into :mod1_vars separated by ' '
		from modelnames;
		select compress(mod2) into :mod2_vars separated by ' '
		from modelnames;
	quit;

	*run proc reg for both models and output RMSE.  The RMSE;
	*is the fitness function and can be changed as per user-needs;
	proc reg data = mydata outest=est (keep=_MODEL_ _RMSE_) noprint;
        	model life_satisfaction= &mod1_vars
			/ selection = &selection sls=&sls sle=&sle  ;
			model life_satisfaction= &mod2_vars
			/ selection = &selection sls=&sls sle=&sle  ;
	run;
	quit;

	*transpose estimate dataset for RMSE comparison.  We want to compare;
	*the RMSE of the two members in the tournament;
	proc transpose data=est out=est;
	run;

	*choose model that wins (smallest RMSE);
	data est (keep = model replicate RMSE);
		set est;
		if col1 < col2 then do;
			model = 1;
			RMSE = col1;
		end;
		else do;
			model=2;
			RMSE = col2;
		end;
		*record which sample it was in tournament;
		replicate = &i;
	run;

	*keep result so we know the winner from each tournament;
	*this dataset keeps the sample number, which of the models won;
	*and the relevant fitness (RMSE);
	proc append base = tournament data = est;
	run;

%end;

%mend;

*call the macro;
%fitness(&reps);

%macro crossmut(vars);
	*now onto the next generation...merge the winners with their ;
	*respective alleles and take best x percent;
	*if this is not the first generation then take a copy of ;
	*nextgen as these are the parents;
	%if %sysfunc(exist(nextgen)) %then %do;
		data oldies;
			set nextgen;
		run;
	%end;

	*if this is not the first run then put the elites back in the mix;
	*the outobs is only necessary for the first run, after this;
	*there will always be half this amount in this table;
	*we are taking out the top performers;
	proc sql outobs= &topcount ;
		create table nextgen as
		select chromosome, RMSE
		from testfit inner join tournament
		on testfit.model = tournament.model and
			testfit.replicate = tournament.replicate
		order by 2 asc;
	quit;

	*add the former elites  into the mix;
	%if %sysfunc(exist(elite)) %then %do;
    	data nextgen;
			set nextgen elite;
		run;
	%end;

	*and the former generations best candidates;
	%if %sysfunc(exist(oldies)) %then %do;
    	data nextgen;
			set nextgen oldies;
		run;
	%end;

	*re-sort and keep best  - keep the new best as elite;
	*if using that option;
	proc sql outobs= &topcount ;
		create table holding as /*use new name to stop warning*/
		select * from nextgen
		order by 2 asc;
	quit;

	*keep the new elites (if using option), this will overwrite the old elites;
	data elite nextgen;
		set holding;
		if _n_ <= &elite then output elite;
		else output nextgen;
	run;

	*now create samples for crossover;
	proc surveyselect data=nextgen method = SRS rep = &crossreps 
                        sampsize = 2 seed = 12345 
						out = crossoverg2 (keep = chromosome replicate) noprint;
  		id _all_;
	run;

	*the code from this point to ***** is the crossover process;
	*start the crossover process, create pieces to swap;
	data crossoverg2 swappiece (keep = replicate model swap);
		retain crosspoint;
		set crossoverg2;
		by replicate;
		*make a random crossover point for each replicate group;
		if first.replicate then crosspoint =int((&vars - 1) * ranuni(35346)) + 1;
		*create the piece to swap;
		swap=substr(chromosome,crosspoint);
		*make counter by replicate group so as to reverse order;
		if first.replicate then model=1; 
		else model = 2;
		output swappiece ;
		output crossoverg2;
	run;

	*sort the pieces in descending order to enable swap;
	proc sort data=swappiece out=swappiece (rename=(swap=swap2));
		by replicate descending model ;
	run;

	*merge back on and do the swap;
	*this step also does the mutation and then outputs the new;
	*chromosome and the replicate number;

	data crossoverg2;
		merge crossoverg2 swappiece;
		*keep a copy of old chromosome for checking;
		oldchro=chromosome;
		*make the new chromosome;
		substr(chromosome,crosspoint)=swap2;
		*now mutate one alele in .1% of cases;
		*first generate a random number between 1 and 10;
		mutrand = int(1000*ranuni(35346)) + 1; 
		*arbitrarily choose 5;
		if mutrand = 5 then do;
			*swap the value using the previously randomly ;
			*generated crosspoint;
			if substr(chromosome,crosspoint,1) = '0' then
				substr(chromosome,crosspoint,1) = '1' ;
			else substr(chromosome,crosspoint,1)='0' ;
		end;
	run;
	*****crossover process ends here;

	*now take the chromosome and turn it back into alleles;
	data testfit (keep = chromosome replicate model allele1-allele&vars);
		array allele { &vars } 3 ;
		set crossoverg2;
		do i = 1 to &vars;
			allele[i]=substr(chromosome,i,1);
		end;
	run;
%mend;

*call the crossover mutation macro;
%crossmut(&outvars);

*terminating function if generations = 500 or avg RMSE doesn't
*change for 20 generations.  These values can be changed as per user needs;
%macro terminate();
*initialise macro variables;
%let fitvalue = 0;
%let laps = 0;
%let end = 0;

*keep going until there have been 20 stable generations or until there;
*have been 500 generations;
%do %until ((%eval(&laps) = &maxit ) or (%eval(&end) = &stable ));
	*store the old fitness average;
	%let fitvalueold = &fitvalue;

	*run the fitness macro;
	%fitness(&crossreps);

	*do the crossover and mutation;
	%crossmut(&outvars);

	proc sql noprint;
		select mean(RMSE) into :fitvalue 
		from nextgen;
	quit;

	*what was the average RMSE of the lap;
%put ****** &fitvalue;


	%if (%sysevalf(&fitvalueold) = %sysevalf(&fitvalue)) 
		%then %let end=%eval(&end+1);
	%else %let end = 0; /*stable counter resets*/
%put ******** stable gens &end;	
	%let laps=%eval(&laps+1);
%put ******** up to lap &laps;
%end;

*write termination point into log;
%put laps= &laps , if &maxit then the maximum iterations were reached;
%put stable = &end;


%mend;

%terminate;

*now print final model details using the winning model;
*first take the top record;
data final (keep = allele1-allele&outvars );
		array allele { &outvars } 3 ;
		set holding;
		if _n_ = 1 ;
			do i = 1 to &outvars;
				allele[i]=substr(chromosome,i,1);
			end;
run;

*transpose and merge with variable names;
proc transpose data=final out=finalturn;
run;

*only keep variable names where the allele (col1 ) is flagged 1;
*store these kept names in variables 'mod1' and 'mod2';
data modelnames;
	merge finalturn var_names;
	if col1 = 1 then mod1 = name;
run;

*first make macro variable from model names again (although this;
*already exists its inside a macro);
proc sql noprint;
	select compress(mod1) into :mod1_vars separated by ' '
	from modelnames;
quit;

*now run the regression;
proc reg data = mydata outest=estfinal tableout  ;
	model life_satisfaction= &mod1_vars
	/ selection =&selection sls= &sls sle= &sle ;
run;
quit;

proc transpose data=estfinal out=estfinal;
id _type_;
run;

*print result to logs;
proc sql noprint;
select compress(_name_)
	into :name separated by '   '
from estfinal
where _name_ ne '_RMSE_' and _name_ ne 'Intercept' 
	and _name_ ne 'Life_Satisfaction' and parms ne .;
quit;

*print to log again;
options notes source source2 errors=20;

*print the selected model variables;
%put Variables kept are: &name;

%let endtime  = %sysfunc(time(),time8.);

*print the run time;
%put run was from &startime to &endtime;


**********end of file**************;


